/*
 *  Copyright (c) 2009 Luca Abeni
 *
 *  This is free software; see gpl-3.0.txt
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>

#include "chunk.h"
#include "input.h"
#include "input-avs.h"

struct input_stream_fixed {
  int fd;
  int csize;
  void *buffer;
};

struct input_stream {
  void *ctx;
  int (*read)(void *s_h, struct chunk *c);
  void (*close)(void *s_h);
  int id;
};

static int chunk_read_fixed(void *s_h, struct chunk *c)
{
  struct input_stream_fixed *s = s_h;
  int res;

  res = read(s->fd, s->buffer, s->csize);

  if (res > 0) {
    c->frame_len[0] = res;
    c->frame[0] = s->buffer;
    c->stride_size = 1;

    return 1;
  }

  return 0;
}

static void input_stream_close_fixed(void *s_h)
{
  struct input_stream_fixed *s = s_h;

  close(s->fd);
  free(s->buffer);
  free(s);
}

static void *input_stream_open_fixed(const char *fname, int chunk_size)
{
  struct input_stream_fixed *s;

  s = malloc(sizeof(struct input_stream_fixed));
  if (s == NULL) {
    return s;
  }
  s->buffer = malloc(chunk_size);
  if (s->buffer == NULL) {
    free(s);

    return NULL;
  }
  s->fd = open(fname, 0);
  if (s->fd < 0) {
    free(s->buffer);
    free(s);

    return NULL;
  }
  s->csize = chunk_size;

  return s;
}

void *input_stream_open(const char *fname, int chunk_size)
{
  struct input_stream *s;

  s = malloc(sizeof(struct input_stream));
  if (s == NULL) {
    return s;
  }
  s->id = 0;
  if (chunk_size <= 0) {
#ifdef AVS
    if (chunk_size == 0) {
        s->read = chunk_read_avs2;
    } else if (chunk_size == -1) {
        s->read = chunk_read_avs1;
    } else {
        s->read = chunk_read_avs2;
    }
    s->close = close_input_stream_avs;
    s->ctx = input_stream_open_avs(fname, chunk_size);
#else
    fprintf(stderr, "Chunk Size <= 0 is unsupported!\n");

    return NULL;
#endif
  } else {
    s->read = chunk_read_fixed;
    s->close = input_stream_close_fixed;
    s->ctx = input_stream_open_fixed(fname, chunk_size);
  }

  return s;
}

int chunk_read(void *s_h, struct chunk *c)
{
  struct input_stream *s = s_h;

  c->id = s->id++;
  return s->read(s->ctx, c);
}


void input_stream_close(void *s_h)
{
  struct input_stream *s = s_h;

  s->close(s->ctx);
}


